/**
 *  @file       log/log.hpp
 *  @author     Taka
 *  @date       23oct97
 *  
 *  @copyright  (c) T. Muraoka 1997 (all rights reserved).
 *              Unauthorized use of this code is prohibited.
 *
 *  @brief      Defines the CMessageLog class.
 * 
 *  @licence    This software is provided 'as-is', without any express 
 *              or implied warranty.
 *
 *              In no event will the author be held liable for any damages 
 *              arising from the use of this software.
 *  
 *              Permission is granted to anyone to use this software 
 *              for any purpose, including commercial applications, 
 *              and to alter it and redistribute it freely, subject to 
 *              the following restrictions: 
 *
 *              - The origin of this software must not be misrepresented; 
 *                you must not claim that you wrote the original software. 
 *                If you use this software in a product, an acknowledgment 
 *                in the product documentation is requested but not required. 
 *
 *              - Altered source versions must be plainly marked as such, 
 *                and must not be misrepresented as being the original software. 
 *                Altered source is encouraged to be submitted back to 
 *                the original author so it can be shared with the community. 
 *                Please share your changes. 
 *
 *              - This notice may not be removed or altered from any 
 *                source distribution. 
 */

#ifndef LOG_LOG_HPP
#define LOG_LOG_HPP

#include <iosfwd>

#include "msgGroup.h"
#include "streambf.h"

namespace debug
{
namespace log
{

class CMessageGroup ; 

/* --- CLASS DEFINITION: CMessageLog ---------------------------------- */

/// Wrapper around an ostream that provides additional formatting services.
/** @class CMessageLog log/log.hpp "log/log.hpp"
 *
 *  This class is the main interface for outputting log messages. If you 
 *  only want a single log, a global instance has been provided that can be
 *  accessed via the theMessageLog() function and macros defined to send
 *  messages to this log. Otherwise, you can create instances of this class,
 *  one for each log, and send messages to them yourself.
 */

class CMessageLog
    : public std::ostream
{

public:                     // --- PUBLIC INTERFACE ---

// enumerations:
    enum 
    {
        DEFAULT_FRAME_TAB_SIZE = 2 ///< Default tab-size to use for log message frames.
    } ;

// constructors:
                            CMessageLog( std::ostream& os ) ;
// destructor:
    virtual                 ~CMessageLog() ;

// message log methods:
                            /// Enables/disables all message groups.
    virtual void            enableMsgLog( bool enable=true ) ; 
                            /// Sets the destination ostream for log messages.
    virtual void            setOutputStream( std::ostream& os ) ;
// message log methods:
                            /// Checks if the message log is enabled.
    bool                    isMsgLogEnabled() const ;

// configuration methods:
                            /// Includes message group ID's in log messages.
    virtual void            enableMsgGroupIDs( bool enable=true ) ;
                            /// Includes message numbers in log messages.
    virtual void            enableMsgNumbers( bool enable=true ) ;
                            /// Includes time-stamps in log messages.
    virtual void            enableTimeStamps( bool enable=true ) ;
                            /// Includes date-stamps in log messages.
    virtual void            enableDateStamps( bool enable=true ) ;
                            /// Includes file-stamps in log messages.
    virtual void            enableFileStamps( bool enable=true ) ;

// message frame methods:
                            /// Starts a new message log frame.
    virtual void            beginFrame() ;
                            /// Ends the current message log frame.
    virtual void            endFrame() ;
// message frame methods:
                            /// Sets the message log frame indentation size.
    void                    setFrameTabSize( int tabSize ) ;

#ifndef _DOXYGEN
// internal methods:
    void                    setState( CMessageGroup* , const char* , size_t ) ;
    CMessageLog&            operator()( CMessageGroup* pMsgGroup , const char* pFilename=NULL , size_t lineNo=0 ) ;
#endif // ! _DOXYGEN

protected:                  // --- PROTECTED INTERFACE ---

// stream methods:
                            /// Processes characters being streamed in.
    virtual int             onStreamBufChar( int ch ) ;
                            /// Output the prefix at the start of each new line.
    virtual void            outputLinePrefix( std::ostream& os ) ;
   
private:                    // --- PRIVATE INTERFACE ---

// copy/assignment (undefined):
                            CMessageLog( const CMessageLog& rhs ) ;
    CMessageLog             operator=( const CMessageLog& rhs ) ;

private:                    // --- DATA MEMBERS ---

// data members:
    bool                    mMsgLogEnabled ; ///< Master enable/disable switch.
// data members:
    bool                    mMsgGroupIDs ; ///< Flags if message group ID's should be included in log messages.
    bool                    mMsgNumbers ; ///< Flags if message numbers should be included in log messages.
    bool                    mTimeStamps ; ///< Flags if time stamps should be included in log messages.
    bool                    mDateStamps ; ///< Flags if date stamps should be included in log messages.
    bool                    mFileStamps ; ///< Flags if filename/line #'s should be included in log messages.
// data members:
    size_t                  mMsgNo ; ///< Current message number.
    CMessageGroup*          mpCurrMsgGroup ; ///< Current message group.
    const char*             mpCurrFileName ; ///< Current file name.
    size_t                  mCurrLineNo ; ///< Current file line number.
// data members:
    int                     mFrameDepth ; ///< Frame nesting depth.
    int                     mFrameTabSize ; ///< Frame tab size.
// data members:
    CMessageLogStreamBuf    mStreamBuf ; ///< Our proxy stream buffer.
    std::streambuf*         mpRealStreamBuf ; ///< The real stream buffer.
    bool                    mIsNewLine ; ///< Flags if we are at the start of a new line.

// friends:
friend class CMessageLogStreamBuf ;

#ifdef _DEBUG
// debugging 
public:
    static bool             mFixedLinePrefix ; ///< Flags if we should output fixed line prefix info.
#endif // _DEBUG

} ;    

/* --- GLOBAL DEFINITIONS --------------------------------------------- */

/// Returns the global CMessageLog.
extern CMessageLog& theMessageLog() ;

/// Installs a new global CMessageLog.
extern CMessageLog* setMessageLog( CMessageLog* pMsgLog ) ;

#ifndef _DOXYGEN
// NOTE: These are use by the LOG_MSG() macros.
extern CMessageLog& theMessageLog( const char* pFilename , size_t lineNo ) ;
extern CMessageLog& theMessageLog( CMessageGroup& msgGroup , const char* pFilename=NULL , size_t lineNo=0 ) ;
#endif // ! _DOXYGEN


} // End namespace log
} // End namespace debug


/* --- MACROS: message logging ---------------------------------------- */

/// @name Message logging macros
/** These macros are provided as a convenience and log messages to the 
 *  global CMessageLog. Compilation can be controlled by defining/un-defining 
 *  the symbol _LOG.
 *
 *  @see theMessageLog()
 */
//@{


#ifdef _DEBUG
#ifdef _LOG
    /** Logs a message (unconditionally). */
    #define LOG_MSG( msg ) \
    { \
	debug::log::theMessageLog(__FILE__,__LINE__) << msg << endl ; \
    }          

    /** Logs a message if the message group is enabled. */
    #define LOG_GMSG( msgGroup , msg ) \
    { \
       debug::log::theMessageLog((msgGroup),__FILE__,__LINE__) << msg << endl ; \
    }          

#else 

    #define LOG_MSG( msg ) { } 
    #define LOG_GMSG( msgGroup , msg ) { } 

#endif //_LOG
#endif //_DEBUG
//@}

/* -------------------------------------------------------------------- */

#endif // LOG_LOG_HPP
