#ifndef BRUSHMGR_H
#define BRUSHMGR_H

#include <xmdv/ARR.h>
#include <xmdv/brush.h>
#include <togl.h>

// Forward declaration of okcdata
struct okcdata;

class BrushMgr
{
	// NOTE: Dimensions in this class's data are not reordered

	// BrushOperMarkDataset
	// When the brush changes, the current brush is pointed

public:
	BrushMgr(okcdata & owner);
	~BrushMgr(void);

	// Reset each brush's counters which keep track of their 
	// coverage. This function should be called after the brushes
	// have been initialized (or otherwise created/reset).
	void ResetBrushCounters(int br);
	void ResetLineColors();

	// Returns true if the line exceeds the boundaries
	inline bool IsLineOutOfBounds(int line) const
	{	return OutOfBounds[line];	}

	// Returns true if the line is inside of the brush br
	inline bool IsLineInBrush(int line, int br) const
	{	return BrushCounters[br][line] <= 0;	}

	// Returns true if the line is inside of the brush operation br_op
	bool IsLineInBrushOper(int line, int br_op) const;

	// Return the number of lines that are not out of bounds.
	inline int GetValidLineCount() const
	{	return ValidLineCount;	}

	// These functions are necessary to keep track of brush coverage
	// when dimensions are hidden, because the BrushMgr should not
	// count dimension d when keeping track of brush coverage.
	// Note that dimensions are in file-space
	void DisableDim(int d);
	void EnableDim(int d);

	// Used for DataValuesText widget in Tcl/Tk script
	int static BrushedDataValuesUpdate(ClientData, Tcl_Interp*, int, char*[]);

	// Used for data summary & statistics
	int GetStats(struct stats_tag & stats, ARR<double> & t_data, int i);

	// One array of values for each brush operation. This is especially
	// useful for the brush data values dialog box.
	ARR< ARR< int > >	BrushOpValues;

	// An array of averages for each brush operation. Averages
	// are not computed if the brush operation does not include
	// an average, or if it is empty.
	ARR< ARR< double> > BrushOpAverages;
protected:
	// The "owner" class which contains BrushMgr
	okcdata & Dataset;

	// An array of bools that marks lines that are out of bounds
	// OutOfBounds[i] == true iff line #i is out of bounds
	// --------------------------------------------------------------
	// Note that this changes the behavior of the application
	// somewhat. Originally, lines that are initially disabled for
	// being out of bounds could become valid if the right dims are
	// hidden. Now, lines that are out of bounds are always disabled.
	ARR<bool> OutOfBounds;
	
	// The number of lines that are not out of bounds
	int ValidLineCount;
	
	// An array of arrays of indices to doubles in okcdata
	// Should have the dimensions [all_dims][data_size]
	// --------------------------------------------------------------
	// For every dim d, Index[d] is a array where each member
	//  is a line number of a line of data. Index[d] is sorted in the
	//  sense that the numbers in Index[d] refer to the lines sorted
	//  by their values at dimension d.
	// --------------------------------------------------------------
	// For every pair of ints i and j:
	//   i <= j holds if and only if, for every dimension d,
	//   LineData[ Index[d][i] ][d] <= LineData[ Index[d][j] ][d]
	//
	// (where LineData[ l ][ d ] is the value of data line l at dim d,
	//  although line data is stored in the awkward okcdata::data_buf)
	ARR< ARR< int > > Index;

	// An array (for each brush ) of arrays (for each line) of the 
	// number of times that each line is outside of each brush.
	// Should have the dimensions [MAXBRUSH][data_size]
	ARR< int > BrushCounters[MAXBRUSH];

	// These functions are used internally to find the range
	// of lines within a brush at a single dimension. The ints
	// returned are indices in Index[dim][x], where x is the
	// value returned.
	int SearchLowerBound(int dim, double v) const;
	int SearchUpperBound(int dim, double v) const;

	// Called by ::BrushedDataValuesUpdate to format a string for a line
	// in the DataValuesText widget.
	void FormatDataValuesLine(int line, char * str);
};

#endif
