////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//      Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //

#include <stdio.h>

#include <xmdv/brush.h>
#include <xmdv/XmdvTool.h>
#include <xmdv/okcdata.h>

//Brush brushes[MAXBRUSH];        // the brushes
//Brush all_brushes[MAXBRUSH];    // the complete brush set
//Brush *g_globals.curr_brush=0;            // the current brush

// Set initial brush position in the middle of a dimension
// and size to be 50 percent of whole range.  Also
// set initial enable/display states.
//
void InitBrushSet()
{
    for (int br=0; br<MAXBRUSH; br++)
    {
		// PAUL: brush_org
		// Get rid of redundant, ugly init code and put it into Brush::InitBrush(...)
		g_globals.brushes[br].Init(br, g_globals.dataset);
		g_globals.all_brushes[br].Copy(g_globals.brushes[br]);
    }

    // Enable the first brush by default
    g_globals.brushes[0].enable = true;
    g_globals.brushes[0].display = true;

    // Set the current brush
    SetActiveBrush(0);
}

// PAUL: brush_org
void Brush::Init(int _num, okcdata * dataset)
{
	// Initialize brushes here instead than in InitBrushSet() above
	num = _num;
	const int all_dims = dataset->all_dims;
	pos.SetLen(all_dims);
	size.SetLen(all_dims);
	bound_offset.SetLen(all_dims);
	min.SetLen(all_dims);
	max.SetLen(all_dims);
	ramp_min.SetLen(all_dims);
	ramp_max.SetLen(all_dims);

    for (int dim=0; dim<all_dims; dim++)
    {

		SetTriple(dim, 
					(g_globals.dataset->dim_max[dim] + g_globals.dataset->dim_min[dim]) * .5,
					(g_globals.dataset->dim_max[dim] - g_globals.dataset->dim_min[dim]) * .5,
					(g_globals.dataset->dim_max[dim] - g_globals.dataset->dim_min[dim]) * .1);
	}
    
	enable = false;
    display = false;
}


int SetActiveBrush(int which)
{
    if (which<0 || which>=MAXBRUSH) return 1;

    g_globals.curr_brush = &g_globals.brushes[which];
    return 0;
}

int SetBrushEnableState(int which, bool state)
{
    if (which<0 || which>=MAXBRUSH) return 1;

    g_globals.brushes[which].enable = state;
    return 0;
}

int SetBrushDisplayState(int which, bool state)
{
    if (which<0 || which>=MAXBRUSH) return 1;

    g_globals.brushes[which].display = state;
    return 0;
}

int SetBrushBoundaryType(int which, Brush::BoundaryType type)
{
    if (which<0 || which>=MAXBRUSH) return 1;

    g_globals.brushes[which].bound_type = type;
    return 0;
}

// Given a data point and a brush number, this
// function checks if the point is covered by the brush
//
bool CheckBrushCoverage(const ARR<double> &point, Brush *brush)
{
	// PAUL:
	// It would be a good idea for people in the future to remove this
	// code since it is well handled by BrushMgr::IsLineInBrush.
	// Unfortunately, "point" may refer to a sequence of values that
	// do not correspond to actual data in the file, so it is left here.

    int i;
    switch(brush->bound_type)
    {
    case Brush::STEP:
        for (i=0; i<g_globals.dataset->dims; i++)
            if (point[i] < brush->Min(i) || point[i] > brush->Max(i))
                return false;
        return true;

    case Brush::RAMP:
        for (i=0; i<g_globals.dataset->dims; i++)
            if (point[i] < brush->RampMin(i) || point[i] > brush->RampMax(i))
                return false;
        return true;
    }
    return false;
}

// Given a data point and a brush number, this
// function checks the amount that the point
// is covered by the brush.  Only makes sense
// for brushes with non stepped boundaries.
// Returns the amount of brush coverage (0..1)
//
double PercentBrushCoverage(const ARR<double> &point, Brush *brush)
{
    switch (brush->bound_type)
    {
    case Brush::STEP: {
        for (int i=0; i<g_globals.dataset->dims; i++)
        {
            if (point[i] < brush->Min(i) || point[i] > brush->Max(i))
                return 0.0;
        }
        return 1.0;
    }

    case Brush::RAMP: {
        // compute coverage as linear interpolation between brush
        // start and end
        //
        double coverage = 1e8;

        int i;
        for (i=0; i<g_globals.dataset->dims; i++)
        {
            double min = brush->Min(i);
            double max = brush->Max(i);
            double offset = brush->BoundOffset(i);

            double dim_coverage = 0;

            if (point[i] >= min && point[i] <= max) {
                // Point is fully in step brush
                dim_coverage = 1.0;
            } else
            if (point[i] >= (min - offset) && point[i] < min) {
                // Point is in lower ramp
                dim_coverage = (offset - (min - point[i])) / offset;
            } else
            if (point[i] <= (max + offset) && point[i] > max) {
                // Point is in upper ramp
                dim_coverage = (offset - (point[i] - max)) / offset;
            } else {
                // No coverage, return 0.0
                return 0.;
            }

            coverage = MIN2(coverage,dim_coverage);
            //coverage += dim_coverage;
        }
        //coverage /= (double)g_globals.dataset->dims;
        return coverage;

    }

    default:
        fprintf(stderr, "Invalid boundary type in PercentBrushCoverage()\n");
    }
    return 0;
}

// Given a data point, returns the number of brushes
// that contain that point.
//
int CoverageRank(const ARR<double> &point)
{
    int coverage=0;
    for (int br=0; br<MAXBRUSH; br++)
    {
        Brush *brush = &g_globals.brushes[br];
        // skip this brush if disabled
        if (brush->enable == false) continue;
        if (CheckBrushCoverage(point, brush)) coverage++;
    }
    return coverage;
}

//
// Check if the brush currently covers this data point
//
bool Brush::IsPointCovered(const ARR<double> &point)
{
    for (int i = 0;i < g_globals.dataset->dims;i++) {
        if (point[i] < min[i] || point[i] > max[i])
            return false;
    }
    return true;
}

//
// Automatic global brush resizing
//
void Brush::Resize(double factor)
{
    if (!g_globals.dataset) return;

    if (factor == 0) {
        // Resize to max
        for (int d=0; d<g_globals.dataset->dims; d++)
		{
			SetPosSize(d, (g_globals.dataset->dim_max[d] + g_globals.dataset->dim_min[d]) / 2.0,
                          g_globals.dataset->dim_max[d] - g_globals.dataset->dim_min[d]);
        }
    }
	else
	{
        for (int d=0; d<g_globals.dataset->dims; d++)
		{
            SetSize(d, size[d] * factor);
        }
    }
}
/*
// PAUL: Remove comments to restore this code
double Brush::Min(int d) 
{
    return pos[d] - size[d]*.5;
}

double Brush::Max(int d) 
{
    return pos[d] + size[d]*.5;
}

double Brush::RampMin(int d) 
{
    return Min(d) - bound_offset[d];
}

double Brush::RampMax(int d) 
{
    return Max(d) + bound_offset[d];
}*/

void Brush::SetMinMax(int d,double lo,double hi)
{
	SetPosSize(d, (hi + lo)*.5, (hi - lo));
}                

// PAUL
// When you set Pos, Size, ...
// The dependent values are automatically updated.
void Brush::SetPos(int d, double _pos)
{
	pos[d] = _pos;
	min[d] = _pos - size[d] * .5;
	max[d] = _pos + size[d] * .5;
	ramp_min[d] = min[d] - bound_offset[d];
	ramp_max[d] = max[d] + bound_offset[d];
}

void Brush::SetSize(int d, double _size)
{
	size[d] = _size;
	min[d] = pos[d] - _size * .5;
	max[d] = pos[d] + _size * .5;
	ramp_min[d] = min[d] - bound_offset[d];
	ramp_max[d] = max[d] + bound_offset[d];
}

void Brush::SetPosSize(int d, double _pos, double _size)
{
	pos[d] = _pos;
	size[d] = _size;
	_size *= .5;
	min[d] = _pos - _size;
	max[d] = _pos + _size;
	ramp_min[d] = min[d] - bound_offset[d];
	ramp_max[d] = max[d] + bound_offset[d];
}

void Brush::SetBoundOffset(int d, double offset)
{
	bound_offset[d] = offset;
	ramp_min[d] = min[d] - offset;
	ramp_max[d] = max[d] + offset;
}

void Brush::SetTriple(int d, double _pos, double _size, double _offset)
{
	pos[d] = _pos;
	size[d] = _size;
	_size *= .5;
	min[d] = _pos - _size;
	max[d] = _pos + _size;
	bound_offset[d] = _offset;
	ramp_min[d] = min[d] - _offset;
	ramp_max[d] = max[d] + _offset;
}

//start_jing_animate_brush
//copy brush to this
void Brush::Copy(Brush & brush)
{
    num = brush.num;
    enable = brush.enable;
    display = brush.display;
    bound_type = brush.bound_type;

    int dim = brush.pos.Len();
    pos.SetLen(dim);
    size.SetLen(dim);
    bound_offset.SetLen(dim);
	// PAUL: brush_org
	// Reset the size of arrays and set values of the new brush
	min.SetLen(dim);
	max.SetLen(dim);
	ramp_min.SetLen(dim);
	ramp_max.SetLen(dim);

    for (int i = 0; i < dim; i++ )
    {
		SetTriple(i, brush.pos[i], brush.size[i], brush.bound_offset[i]);
    }
}
//end_jing_animate_brush


