////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  XmdvTool Version 4.0 - A program for Multivariate Data Display
//
//  Copyright 1994-1999, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//  	Matthew O. Ward, Allen R. Martin, Ying-Huey Fua
//  	Computer Science Department
//  	Worcester Polytechnic Institute
//  	Worcester, MA 01609 USA
//
//  	http://davis.wpi.edu/~xmdv
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  XmdvTool Development Team: <matt@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////// COPYRIGHT //

// PROJECT INCLUDES
//
#include <xmdv/colormap.h>
#include <xmdv/ogl.h>
#include <xmdv/colorassign.h>
#include <xmdv/colorramp.h>
#include <xmdv/Image.h>
#include <xmdv/ColorMapEditor.h>
#include <xmdv/XmdvTool.h>

//-------------------------------------------------------------------------
colormap::colormap(const std::string &name, ColorRamp *ramp /*= NULL*/, 
				   int len /*= CMAPSIZE*/, ColorModel model /*= MODEL_RGB*/)
{
	m_name = name;
	m_len = len;
	m_model = model;
	m_colorramp = NULL;
	if (ramp) {
		SetColorRamp(ramp);
	} else {
		SetColorRamp(ColorRamp::GetPredefined2(len));
	}
	m_checkImg = NULL;
	g_globals.RegisterCMap(*this);
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
void colormap::InitTex()
{
	m_checkImg = new Image(64,64);
	m_checkImg->Clear(255,255,255);
	int i, j;
	unsigned char c;

	for (i = 0; i < 64; i++) {
		for (j = 0; j < 64; j++) {
			if ( ((i & 0x8) == 0) ^ ((j & 0x8) == 0) ) { 
				c = (unsigned char)255;
			} else { 
				c = (unsigned char)127; 
			}
			m_checkImg->WritePixel(i,j,c,c,c);
		}
	}
	m_checkImg->CreateTexture();
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
colormap::colormap(const colormap& map)
		: m_colorramp(new ColorRamp(*map.m_colorramp))
{
	m_name = map.m_name;
	m_len = map.m_len;
	m_model = map.m_model;
	if (map.m_checkImg) {
		InitTex();
	} else {
		m_checkImg = NULL;
	}
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
void colormap::operator = (const colormap& map)
{
	if (&map == this) 
		return;

	m_name = map.m_name;
	m_len = map.m_len;
	m_model = map.m_model;
	delete m_colorramp;
	m_colorramp = new ColorRamp(*map.m_colorramp);
	if (map.m_checkImg) {
		InitTex();
	} else {
		m_checkImg = NULL;
	}
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
/*virtual*/ colormap::~colormap()
{
	g_globals.UnregisterCMap(*this);
	delete m_colorramp;
	if (m_checkImg) delete m_checkImg;
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
void colormap::SetColorRamp(ColorRamp *ramp)
{
	if (m_colorramp)
		delete m_colorramp;
	m_colorramp = ramp;
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
void colormap::SetColorModel(colormap::ColorModel model)
{
	m_model = model;
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
unsigned long colormap::operator[] (int index) const
{
	if (m_model == MODEL_RGB) {
		return (
		((unsigned long) (255.0f * m_colorramp->Evaluate(index, 2))) |
		(((unsigned long) (255.0f * m_colorramp->Evaluate(index, 1))) << 8) |
		(((unsigned long) (255.0f * m_colorramp->Evaluate(index, 0))) << 16) |
		(((unsigned long) (255.0f * m_colorramp->Evaluate(index, 3))) << 24));
	} else if (m_model == MODEL_GREY) {
		return (
			((unsigned long) (255.0f * m_colorramp->Evaluate(index, 0))) |
			(((unsigned long) (255.0f * m_colorramp->Evaluate(index, 0))) << 8) |
			(((unsigned long) (255.0f * m_colorramp->Evaluate(index, 0))) << 16) |
			(((unsigned long) (255.0f * m_colorramp->Evaluate(index, 3))) << 24));
	} else if (m_model == MODEL_HSV) {
		return HSIColor(6.0f * m_colorramp->Evaluate(index, 0),
						m_colorramp->Evaluate(index, 1),
						m_colorramp->Evaluate(index, 2),
						m_colorramp->Evaluate(index, 3)).getRGB();
	}

	return 0;
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
RGBt colormap::GetRGBAt(int index) const
{
	if (m_model == MODEL_RGB) {
		return RGBt(
			((unsigned long) (255.0f * m_colorramp->Evaluate(index, 2))) |
			(((unsigned long) (255.0f * m_colorramp->Evaluate(index, 1))) << 8) |
			(((unsigned long) (255.0f * m_colorramp->Evaluate(index, 0))) << 16) |
			(((unsigned long) (255.0f * m_colorramp->Evaluate(index, 3))) << 24));
	} else if (m_model == MODEL_GREY) {
		return (
			((unsigned long) (255.0f * m_colorramp->Evaluate(index, 0))) |
			(((unsigned long) (255.0f * m_colorramp->Evaluate(index, 0))) << 8) |
			(((unsigned long) (255.0f * m_colorramp->Evaluate(index, 0))) << 16) |
			(((unsigned long) (255.0f * m_colorramp->Evaluate(index, 3))) << 24));
	} else if (m_model == MODEL_HSV) {
		return RGBt(HSIColor(6.0f * m_colorramp->Evaluate(index, 0),
			m_colorramp->Evaluate(index, 1),
			m_colorramp->Evaluate(index, 2),
			m_colorramp->Evaluate(index, 3)).getRGB());
	}

	return 0;
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
HSIColor colormap::GetHSIAt(int index) const
{
	if (m_model == MODEL_RGB) {
		return HSIColor(
			((unsigned long) (255.0f * m_colorramp->Evaluate(index, 2))) |
			(((unsigned long) (255.0f * m_colorramp->Evaluate(index, 1))) << 8) |
			(((unsigned long) (255.0f * m_colorramp->Evaluate(index, 0))) << 16) |
			(((unsigned long) (255.0f * m_colorramp->Evaluate(index, 3))) << 24));
	} else if (m_model == MODEL_GREY) {
			return HSIColor(
				((unsigned long) (255.0f * m_colorramp->Evaluate(index, 0))) |
				(((unsigned long) (255.0f * m_colorramp->Evaluate(index, 0))) << 8) |
				(((unsigned long) (255.0f * m_colorramp->Evaluate(index, 0))) << 16) |
				(((unsigned long) (255.0f * m_colorramp->Evaluate(index, 3))) << 24));
	} else if (m_model == MODEL_HSV) {
		return HSIColor(6.0f * m_colorramp->Evaluate(index, 0),
						m_colorramp->Evaluate(index, 1),
						m_colorramp->Evaluate(index, 2),
						m_colorramp->Evaluate(index, 3));
	}

	return 0;
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
void colormap::DrawColorBar(float x1 /*= 0.0f*/, float y1 /*= 0.0f*/,
							float x2 /*= 1.0f*/, float y2 /*= 0.0f*/,
							float width /*= 5.0f*/)
{
	glPushMatrix();

	// Draw the textured quad first
	if (m_checkImg) {
		glPushMatrix();
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glTranslatef(x1+1.0f,y1-width+1.0f,0);
		glScalef(fabs(x2-x1),width+3.0f, 1.0f);
		m_checkImg->EnableTexturing();
		glBegin(GL_QUADS); {
			glTexCoord2f(0.0f, 0.0f); glVertex2f(0.0f, 0.0f);
			glTexCoord2f(3.0f, 0.0f); glVertex2f(1.0f, 0.0f);
			glTexCoord2f(3.0f, 1.0f); glVertex2f(1.0f, 1.0f);
			glTexCoord2f(0.0f, 1.0f); glVertex2f(0.0f, 1.0f);
		}
		glEnd();
		m_checkImg->DisableTexturing();
		glPopMatrix();
	}
	// Draw the map

	glPushMatrix();
	glShadeModel(GL_SMOOTH);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glLineWidth(width);
	glBegin(GL_LINE_STRIP); {
		// initial x and y
		double x = x1, y = y1;
		// deltas
		double dx = x2 - x1, dy = y2 - y1;
		// step size
		double step = 1 / double(CMAPSIZE - 1);
		// for each delta step, draw a line of the 
		// specified width with the color in the lut
		for (int i = 0; i < CMAPSIZE; i++) {
			unsigned long color = (*this)[i];
			glColor4ub((unsigned char) (color >> 16) & 0xff,
					(unsigned char) (color >> 8) & 0xff,
					(unsigned char) (color >> 0) & 0xff,
					(unsigned char) (color >> 24) & 0xff);
			x += step*dx; y += step*dy;
			glVertex2f((float)x, (float)y);
		}
	}
	glEnd();

	glLineWidth(1.0);
	glShadeModel(GL_FLAT);
	glDisable(GL_BLEND);
	glPopMatrix();

	glPopMatrix();
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
/*static*/ colormap *colormap::Load(std::istream &inputstream, 
									const std::string &cmapname)
{

	return NULL;
}
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
/*static*/ bool	colormap::Save(colormap *cmap, std::ostream &outputstream)
{
	return false;
}
//-------------------------------------------------------------------------
