#include <Cache.h>
#include <Brush.h>
#include <CacheSearcher.h>
#include <ICacheEntry.h>

using namespace cache;
/**
* Creates the sequential index structure
* \param sizeOfCache 
* \return 
*/
SequentialIndex::SequentialIndex(unsigned long sizeOfCache):seqList(sizeOfCache)
{
	this->sizeOfCache = sizeOfCache;
	for(unsigned long i=0;i<sizeOfCache;i++)
	{
		seqList[i] = NULL;
	}	
}

void SequentialIndex::deleteData(Tools::Geometry::IShape *iShape,
							     unsigned long id)
{
	//Just simply make this point to NULL;
	seqList[id] = NULL;
}

/**
* Inserts the cache entry into the index.
* Called when a new 
* \param &cacheEntry 
* \return 
*/
void SequentialIndex::insertData(ICacheEntry& cacheEntry)
{
	this->seqList[cacheEntry.getID()] = &cacheEntry;
}

/**
* Does a sequential scan over the complete data set in the main memory
* \param &brush 
* \return 
*/
std::list<ICacheEntry*>& SequentialIndex ::search(Brush &brush)
{
	this->resultList.clear();
	double low[2],high[2];
	low[0] = 0.0;
	low[1] = 0.25;
	high[0] = 0.6;
	high[1] = 0.75;
	Tools::Geometry::Region dummyRegion(low,high,2);
	for(unsigned long i=0;i<sizeOfCache;i++)
	{
		if(seqList[i] != NULL)
		{		
			if(brush.inBrush(*seqList[i]->getIShape()))
				this->resultList.push_back(seqList[i]);
		} // for(unsigned long i=0;i<sizeOfCache;i++)		
		else
		{
			brush.inBrush(dummyRegion);//Hack to see if the inBrush operation affects the
			//sequential scan as much., this will work only initially
			//so if the user traces are not very long the R-Tree should
			//still win.
		}
	}
	return(this->resultList);
}

SequentialIndex::~SequentialIndex()
{}

RTreeIndex::RTreeIndex(double fillFactor, 
						unsigned long indexCapacity,			  
						unsigned long leafCapacity,
						unsigned numberOfDimensions):
						cacheSearcher(new CacheSearcher(&resultList))
{
	this->indexIdentifier  = 1;
	this->memoryStorageManager = 
		SpatialIndex::StorageManager::createNewMemoryStorageManager();			
	this->rTree = 
		SpatialIndex::RTree::createNewRTree(*memoryStorageManager,
		fillFactor,
		indexCapacity,
		leafCapacity,
		numberOfDimensions,
		SpatialIndex::RTree::RV_RSTAR,
		indexIdentifier);				
}

ICacheEntry* RTreeIndex::findFarthestNode(Brush *currentBrush)
{
	this->rTree->queryStrategy(*this);
	return(NULL);
}

void RTreeIndex::getNextEntry(const SpatialIndex::IEntry& previouslyFetched, 
								long& nextEntryToFetch, 
								bool& bFetchNextEntry)
{
	
	//NodePtr root = this->rTree->readNode(m_rootID);
	//Check to see if node is a child node
	//Note here I assume that object is 2 dimensional, 

	/*stack<NodePtr> st;
	NodePtr root = readNode(m_rootID);

	if (root->m_children > 0 && query.intersectsShape(root->m_nodeMBR)) st.push(root);

	while (! st.empty())
	{
		NodePtr n = st.top(); st.pop();

		if (n->m_level == 0)
		{
			v.visitNode(*n);

			for (unsigned long cChild = 0; cChild < n->m_children; cChild++)
			{
				bool b = false;
				switch(type)
				{
				case ContainmentQuery : 
					b = query.containsShape(*(n->m_ptrMBR[cChild]));
					break;
				case IntersectionQuery:
					b = query.intersectsShape(*(n->m_ptrMBR[cChild]));
					break;
				case QueryContainment:														
					b = (*(n->m_ptrMBR[cChild]))
						.containsShape(query);							
					break;						

				}					

				if (b)
				{
					Data data = Data(n->m_pDataLength[cChild], n->m_pData[cChild], *(n->m_ptrMBR[cChild]), n->m_pIdentifier[cChild]);
					v.visitData(data);
					m_stats.m_queryResults++;
				}
			}
		}
		else
		{
			v.visitNode(*n);
			for (unsigned long cChild = 0; cChild < n->m_children; cChild++)
			{
				if (query.intersectsShape(*(n->m_ptrMBR[cChild]))) st.push(readNode(n->m_pIdentifier[cChild]));
			}
		}
	}*/
}


/**
* Inserts the cache entry into the index.
* Called when a new 
* \param &cacheEntry 
* \return 
*/
void RTreeIndex::insertData(ICacheEntry& cacheEntry)
{	
	//Note with each entry in the R Tree we store the 
	//1) The address of the cacheEntry it is pointing to.
	ICacheEntry* address = &cacheEntry;
	//We want to make the address a byte array
	//The base of the byte array should be the same as address(CacheEntry pointer)
	byte *pData = reinterpret_cast<byte*> (&address);

	this->rTree->insertData(sizeof(ICacheEntry*),
		pData,
		*cacheEntry.getIShape(),
		cacheEntry.getID());
}

/**
* Deletes the dataitem from the index
* \param iShape 
* \param id 
* \return 
*/
void RTreeIndex::deleteData(Tools::Geometry::IShape* iShape,unsigned long id)
{
	this->rTree->deleteData(*iShape,id);	
}

/**
* Deletes the data item from the br
* \param brush 
* \param resultList 
* \return 
*/
std::list<ICacheEntry*>& RTreeIndex::search(cache::Brush& brush)
{
	//Clear the list to avoid cumulative effect
	this->resultList.clear();

	Tools::Geometry::IShape& iShape =brush.getShape();		

	//the cache searcher implements the visitor and creates the result List
	this->cacheSearcher->setSearchParameters(&brush);
	this->rTree->intersectsWithQuery
		(brush.getShape(),
		*(this->cacheSearcher));		

	return(this->resultList);
}


/**
* Destroys the R Tree index structure
* \return 
*/
RTreeIndex::~RTreeIndex()
{
	delete(rTree);
	delete(memoryStorageManager);
	delete(this->cacheSearcher);
}


