#include <cache/CacheBuilder.h>
#include <CacheManager.h>
#include <iostream>
#include <LRUReplacement.h>
#include <ReplacementPolicy.h>
#include <CacheIterator.h>
#include <CacheRegion.h>
#include <Loader.h>	
#include <DataBaseLibrary/ConnectionFactory.h>
#include <DataBaseLibrary/Connection.h>
#include <DataBaseLibrary/Statement.h>
#include <PrefetchController.h>
#include <Prefetcher.h>
#include <sstream>
#include <BrushStore.h>
#include <zthread/Thread.h>
#include <StatisticsGatherer.h>
#include <fstream>
#include <Timer/Timer.h>
#include <ConfigurationStore.h>
#include <ICacheController.h>
#include <CacheController.h>
#include <DataIterator.h>
#include <CacheEntry.h>
#include <Cache.h>
#include <LoadRequester.h>
#include <QueryFactory.h>
#include <log/log.h>
#include <cache/RegionFactory.h>
#include <cache/SuperIndex.h>

using namespace cache;

/*
*	A Factory class that 
*  builds the cache connects the different componets together
*  It can build the cache for the following 3 configurations
* 1) No Prefetcher
* 2) Fixed Prefetcher
* 3) Adaptive Prefetcher
*/
/*
/*The Data base connection parameters
* username
* passwd
* dsn source
*/
/*	The input parameters the class requires for Cache
* 1) Cache Size 
* 2) fillFactor = 0.7; //Default Value = 0.7
* 3) indexCapacity = 2000; //Need to have some default value
* 4) leafCapacity =  2000; //Need to have some default value
* 5) Replacement policy to use 
* ============================================================
* Prefetcher
* 1) Prefetcher Enabled or not
* 2) Prefetch adaptive or fixed
* 3) Initial Prefetch strategy to be used. //Default value should be there
* 3) brush store size //Default value is 2
* 
*/
/*

*/

/**
* Makes the new Prefetcher
*/
template <class Policy,class Value>
void CacheBuilder::Builder<Policy,Value>::makePrefetcher()
{
	int adaptivePrefetching = 0;
	if(this->cacheConfigStore->getPrefetcherType() == 
		CacheConfigStore::Adaptive)
	{
		this->statisticsGatherer = new StatisticsGatherer();
		adaptivePrefetching = 1;
	}

	this->prefetcher = new Prefetcher(cacheConfigStore->getCacheSize(),
										100, //Number of levels				
										this->brushStore,
										adaptivePrefetching,
										this->statisticsGatherer,
										cacheConfigStore->getPrefetchStrategy(),
										11);//Number of P entries

	this->prefetchController = new PrefetchController(this->prefetcher);							
}


/**
* Returns the Cache controller
* \return 
*/
ICacheController* CacheBuilder::getCacheController()	
{
	ICacheController* cacheController;
	if(bLRU != NULL)
		cacheController = this->bLRU->cacheController;

	if(bDistance != NULL)
		cacheController = this->bDistance->cacheController;

	return(cacheController);
}


template <class Policy,class Value>
CacheBuilder::Builder<Policy,Value>::~Builder()
{
	if(this->prefetchController != NULL)
		delete(prefetchController);
	if(this->cacheController != NULL)
		delete(cacheController);
	if(this->cacheManager != NULL)
		delete(cacheManager);
	if(this->brushStore != NULL)
		delete(brushStore);
	if(this->statisticsGatherer != NULL)
		delete(statisticsGatherer);
	if(this->loadRequester != NULL)
		delete(loadRequester);
	if(this->queryFactory != NULL)
		delete(this->queryFactory);
	if(this->perStatistics != NULL)
		delete(perStatistics);
	QuerySimpleFactory::getReference()= NULL;
	QueryGeomFactory::getReference() = NULL;
	delete(AscRegionFactory::getInstance());
	delete(DescRegionFactory::getInstance());
	AscRegionFactory::getReference() = NULL;	
	DescRegionFactory::getReference() = NULL;
	
	delete(statement);
	delete(connection);
	delete(loader);
	delete(index);
	delete(schema);
}

/*
 *	Returns the PrefetchController.
 */
cache::PrefetchController* cache::CacheBuilder::getPrefetchController()
{
	PrefetchController* prefetchController = NULL;

	if(this->bLRU != NULL)
		prefetchController = this->bLRU->getPrefetchController();
	if(this->bDistance != NULL)
		prefetchController = this->bDistance->getPrefetchController();

	return(prefetchController);
}


cache::CacheBuilder::CacheBuilder(CacheConfigStore* cacheConfigStore,
								  DBConfigStore* dbConfigStore):
								  bLRU(NULL),bDistance(NULL)
{	
	// if(cacheConfigStore->getReplacementPolicy() == "LRU")	
	//Add a if statement for each of the new replacement policies.
	if(cacheConfigStore->getReplacementPolicy() == "LRU")
	{
		bLRU = new Builder<LRUReplacement,LRUReplacementValue>
							(cacheConfigStore,dbConfigStore);
	}
	if(cacheConfigStore->getReplacementPolicy() == "Distance")
	{
		bDistance = new Builder<DistanceReplacement,bool>
			(cacheConfigStore,dbConfigStore);
	}
}

cache::CacheBuilder::CacheBuilder():bLRU(NULL),bDistance(NULL)
{
}

cache::CacheBuilder::~CacheBuilder()
{
	if(bLRU != NULL)
		delete(this->bLRU);
	if(bDistance != NULL)
		delete(this->bDistance);
}

/*
 *	Makes a new cache manager
 */
template <class Policy,class Value>
void CacheBuilder::Builder<Policy,Value>::makeCacheManager()
{
	//First make the connection
	connection = 
		::ConnectionFactory::getInstance()->makeConnection(
		dbConfigStore->userName,
		dbConfigStore->passwd,
		dbConfigStore->dsnsource);

	ICacheEntry* buffer = makeBuffer();
	if(cacheConfigStore->memoryIndex == "SuperIndex")
	{		
		RTreeIndex* rtreeIndex = new RTreeIndex(cacheConfigStore->getRTreeFillFactor(),
			cacheConfigStore->getRTreeIndexCapacity(),
			cacheConfigStore->getRTreeLeafCapacity(),
			2);

		SequentialIndex* seqIndex = new SequentialIndex(cacheConfigStore->cacheSize);


		this->index = new SuperIndex(cacheConfigStore->cacheSize,
									 2,
									 this->perStatistics,
									 seqIndex,
									 rtreeIndex);
	
	} // if(cacheConfigStore->memoryIndex == SuperIndex)
	else if(cacheConfigStore->memoryIndex == "RTree")
	{
		this->index = new RTreeIndex(cacheConfigStore->getRTreeFillFactor(),
			cacheConfigStore->getRTreeIndexCapacity(),
			cacheConfigStore->getRTreeLeafCapacity(),
			2);
	}
	else
	{				
		this->index = new SequentialIndex(cacheConfigStore->getCacheSize());
	}

	//Instantiated the cache
	this->cache = new Cache(cacheConfigStore->getCacheSize(),					
		buffer,
		index);

	//Makes a new replacement policy
	this->replacementPolicy = makeReplacementPolicy(buffer);

	//Instantiate the statement												 dsnsource);
	this->statement = connection->createStatement();


	//Copying the schema from the dbConfigStore.
	//Using the copy constructor to copy the schema
	this->schema = new Schema(*dbConfigStore->schema);

	//Instantiate the loader
	this->loader = new Loader(statement,this->schema);				

	//Make the appropriate query factory
	if(cacheConfigStore->queryType == "Geom")
		this->queryFactory = QueryGeomFactory::makeQueryFactory(this->schema,
		dbConfigStore->tableName);
	else
		this->queryFactory = QuerySimpleFactory::makeQueryFactory(this->schema,
		dbConfigStore->tableName);	


	//Given the cache configuration parameters you now have to 
	//build the cache here.
	this->cacheManager = new CacheManager(this->loader,
		this->cache,
		this->replacementPolicy,
		this->queryFactory,
		this->cacheConfigStore->fragmentationThreshold);			

}


template <class Policy,class Value>
CacheBuilder::Builder<Policy,Value>::Builder(CacheConfigStore* cacheConfigStore,
											 DBConfigStore* dbConfigStore)
											 :cacheConfigStore(cacheConfigStore),dbConfigStore(dbConfigStore)			
{
	//Initializing all The pointers to NULL			
	prefetchController = NULL;			
	prefetcher = NULL;			
	cacheController = NULL;			
	cacheManager = NULL;			
	brushStore = NULL;			
	statisticsGatherer= NULL;
	loadRequester = NULL;
	this->queryFactory = NULL;
	this->perStatistics = NULL;

	////////////////////////////////////////
	this->perStatistics = new PerformanceStatistics();
	//The cache manager uses the performance statistics.
	makeCacheManager();

	this->brushStore = 
		new BrushStore(cacheConfigStore->getSizeOfBrushStore());

	if(cacheConfigStore->getPrefetchStrategy() != 
		CacheConfigStore::Noprefetch)
	{		
		//1)This instantiates the statistics gatherer
		//2)The prefetcher
		//3)The prefetch controller
		makePrefetcher();
		LoadRequester:: make(this->cacheManager,
			this->prefetchController);
		this->loadRequester = LoadRequester::getInstance();
	}
		
	this->cacheController = new 
		CacheController
		(this->cacheManager,
		2,
		this->brushStore,
		this->perStatistics,
		this->prefetchController);
		

	if(this->cacheConfigStore->getPrefetcherType() == 
		CacheConfigStore::Adaptive)
	{			
		this->cacheController->addLoadListener(this->statisticsGatherer);
		this->cacheManager->addServedItemListener(this->statisticsGatherer);
	}
}
