/* file: strategyTable.c 
 * purpose: define strategyTable class
 * revision log:
 */
#include <strategyTable.h>
#include <globalAverage.h>
#include <expAverage.h>
#include <stdlib.h>
#include <stdio.h>
#include <cstdlib>    /* for srand() and rand() */
#include <fstream>

// P120202 Cost factors for misclassification cost
#define		C_NP	0.5		/* Cost for not predicting */
#define		C_MP	0.5		/* Cost for mis-predicting */

//extern void outFileDebug(char *);

#define RAND01()    ((float) rand()/RAND_MAX)   
 /* Return a uniform random number in the range [0,1].
  * This method is internal to this class.
  */
using namespace std;

struct strategyTable_Str
 /* Defines 'strategyTable_Str' as a pointer to a structure that
  * is not defined here and so is not accessible to the outside world
  */
{	int nStrategies ;  /* # strategies in the table */
	int strategyId[STRATEGYTABLE_MAXSTRATEGIES] ;  
	globalAverage globAveRespTime[STRATEGYTABLE_MAXSTRATEGIES] ; /* global ave response time */
	expAverage expAveRespTime[STRATEGYTABLE_MAXSTRATEGIES] ;  /* exp smoothed ave response time */
	int nSelected[STRATEGYTABLE_MAXSTRATEGIES] ;   /* # times selected */

	// Punit120202 for  new fitness function
	int correctlyPredicted[STRATEGYTABLE_MAXSTRATEGIES] ; /* total # of correctly predicted objects */
	int notPredicted[STRATEGYTABLE_MAXSTRATEGIES] ; /* total # of objects that are not predicted */
	int misPredicted[STRATEGYTABLE_MAXSTRATEGIES] ; /* total # of objects that are mis-predicted */
	expAverage expAveMisClassificationCost[STRATEGYTABLE_MAXSTRATEGIES] ; /* exp smoothed ave mis-classification rate */
	globalAverage globalAveMisClassificationCost[STRATEGYTABLE_MAXSTRATEGIES] ; /* exp smoothed ave mis-classification rate */

	globalAverage overallGlobAveRespTime ; /* global ave response time across all strategies */
	expAverage overallExpAveRespTime ; /* exp smoothed ave response time across all strategies */

	int overallCorrectlyPredicted ;
	int overallNotPredicted ;
	int overallMisPredicted ;
	globalAverage overallGlobAveMisClassCost ; /* global ave misclassification cost across all strategies */
	expAverage overallExpAveMisClassCost ; /* exp smoothed ave misclassification cost across all strategies */

	int overallNSelected ;  /* sum of nSelected entries */

	int selectedStrategyId ;  /* current active strtegy id */
	
	int selectionPolicy ;	/* id for strategy selection policy */
	int fitnessFunction ;   /* id for fitness function */
	int randSeed ;     /* seed for [0,1] uniform random number generator */
} strategyTable_Str ;

typedef struct strategyTable_Str *strategyTable ;
 /* Defines 'strategyTable_Str' as a pointer to a structure that
  * is not defined here and so is not accessible to the outside world
  */

strategyTable t ;


void strategyTable_Create()
 /*	Returns a new strategyTable, or NULL if there isn't enough memory.
  */
{
	//outFileDebug("BEGIN strategyTable_Create()\n" );

	t = (strategyTable) malloc(sizeof(struct strategyTable_Str)) ;
	if (t != NULL)
	{ 	/* initialize for overall statistics */
		t->overallGlobAveRespTime = globalAverage_Create() ;
		t->overallExpAveRespTime  = expAverage_Create() ;

		t->overallGlobAveMisClassCost = globalAverage_Create() ;
		t->overallExpAveMisClassCost  = expAverage_Create() ;
		t->overallCorrectlyPredicted = 0 ;
		t->overallMisPredicted = 0 ;
		t->overallNotPredicted = 0 ;
		t->overallNSelected = 0 ;
		t->selectedStrategyId = NO_PREFETCH ;

		t->nStrategies = 0 ;  

		t->selectionPolicy = STRATEGYTABLE_POLICYBEST ;  /* default selection policy */
		t->fitnessFunction = STRATEGYTABLE_FITGLOBAVERESPTIME ; /* default fitness function */

		/* initialize the random number generator */
		srand(STRATEGYTABLE_RANDSEED) ;
	}

	//outFileDebug("END strategyTable_Create()\n" );
}

void strategyTable_Destroy(void) 
 /* Frees the memory associated with 'strategyTable'. 
  */
{ 
	//outFileDebug("BEGIN strategyTable_Destroy()\n" );

	// destroy each complex array
	for (int i = 0 ; i < t->nStrategies; i++)
	{
		globalAverage_Destroy(t->globAveRespTime[i]) ;
		expAverage_Destroy(t->expAveRespTime[i]) ;

		globalAverage_Destroy(t->globalAveMisClassificationCost[i]) ;
		expAverage_Destroy(t->expAveMisClassificationCost[i]) ;
	}

	// destroy overall statistics
	globalAverage_Destroy(t->overallGlobAveRespTime) ;
	expAverage_Destroy(t->overallExpAveRespTime) ;

	globalAverage_Destroy(t->overallGlobAveMisClassCost) ;
	expAverage_Destroy(t->overallExpAveMisClassCost) ;

	free(t) ;

	//outFileDebug("END strategyTable_Destroy()\n" );
}

float strategyTable_calculateMisClassCost(int correctlyPredicted, int notPredicted, int misPredicted)
 /* Finds the Mis-Classification Cost for the given values
  */
{
	float percentNotPredicted =0 ;
	float percentMisPredicted =0 ;

	int total = correctlyPredicted + notPredicted + misPredicted ;
//	int total = correctlyPredicted + notPredicted ;

	if (total == 0)	total = 1 ;
	percentNotPredicted = (float) notPredicted/(total) ;
	percentMisPredicted = (float) misPredicted/(total) ;

	return (float) ( (C_NP * percentNotPredicted) + (C_MP * percentMisPredicted) );
}

void strategyTable_SetSelectionPolicy(int policy)
 /* Sets the Strategy Selection Policy
  */
{
	//outFileDebug("BEGIN strategyTable_SetSelectionPolicy()\n" );

	if (policy >= 0 && policy < STRATEGYTABLE_NPOLICIES)
		t->selectionPolicy = policy ;
	else
	{
		ofstream fp(ERRFILE, ios::app);
		fp << "strategyTable_SetSelectionPolicy() Error: invalid policy code " << policy << endl ;
		exit(1);
	}

	//outFileDebug("END strategyTable_SetSelectionPolicy()\n" );
}


void strategyTable_SetFitnessFunction(int fitnessFunction)
 /* Sets the Fitness Function
  */
{
	//outFileDebug("BEGIN strategyTable_SetFitnessFunction()\n" );
	if (fitnessFunction >= 0 && fitnessFunction < STRATEGYTABLE_NFITNESS)
		t->fitnessFunction = fitnessFunction ;
	else
	{
		ofstream fp(ERRFILE, ios::app);
		fp << "strategyTable_SetFitnessFunction() Error: invalid fitness function code " << fitnessFunction << endl ;
		exit(1);
	}
	//outFileDebug("END strategyTable_SetFitnessFunction()\n" );
}

void strategyTable_AddStrategy(int strategyId) 
 /* Adds a strategy to the table
  */
{
	//outFileDebug("BEGIN strategyTable_AddStrategy()\n" );

	if (t->nStrategies < STRATEGYTABLE_MAXSTRATEGIES)
	{
		t->strategyId[t->nStrategies] = strategyId ;
		t->globAveRespTime[t->nStrategies] = globalAverage_Create() ;  //??check null
		t->expAveRespTime[t->nStrategies] = expAverage_Create() ;

		// P120302 Adding the methods
		t->expAveMisClassificationCost[t->nStrategies] = expAverage_Create() ;
		t->globalAveMisClassificationCost[t->nStrategies] = globalAverage_Create() ;

	    // initialize all global and exp average response times to 0, the best response time. 
		//?? initialize to best? or initialize to average?? but what is average??x
		globalAverage_Update(t->globAveRespTime[t->nStrategies], 0.0) ;
		expAverage_Update(t->expAveRespTime[t->nStrategies], 0.0) ;

		// P120202
		t->correctlyPredicted[t->nStrategies] = 0 ;
		t->misPredicted[t->nStrategies] = 0 ;
		t->notPredicted[t->nStrategies] = 0 ;

		t->nSelected[t->nStrategies] = 1 ;

		globalAverage_Update(t->overallGlobAveRespTime, 0.0) ;
		expAverage_Update(t->overallExpAveRespTime, 0.0) ;

		globalAverage_Update(t->overallGlobAveMisClassCost, 0.0) ;
		expAverage_Update(t->overallExpAveMisClassCost, 0.0) ;

		t->overallNSelected++ ;
		t->nStrategies++ ;
	}
	else
	{
		ofstream fp(ERRFILE, ios::app);
		fp << "strategyTable_AddStrategy() Error: invalid strategy id " << strategyId << endl ;
		exit(1);
	}

	//outFileDebug("END strategyTable_AddStrategy()\n" );
}


void strategyTable_Print(void)
 /* Prints contents of the strategy table
  */
{
	char message[126];
	FILE *fp = fopen(OUTPUTFILE, "a");
	//outFileDebug("BEGIN strategyTable_Print()\n" );
	//outFileDebug("\n Id    GlAvRT     ExAvRT        CP      NP      MP    ExMisCl    GlMisCl      #Selected\n");
	fprintf(fp, "\n Id    GlAvRT     ExAvRT        CP      NP      MP    ExMisCl    GlMisCl      #Selected\n");

	for (int i = 0; i < t->nStrategies; i++)
	{
		sprintf(message, "%3d   %8.4f   %8.4f   %5d   %5d   %5d   %8.4f   %8.4f   %5d\n", t->strategyId[i], 
			globalAverage_GetValue(t->globAveRespTime[i]), expAverage_GetValue(t->expAveRespTime[i]), 
			t->correctlyPredicted[i], t->notPredicted[i], t->misPredicted[i], 
			expAverage_GetValue(t->expAveMisClassificationCost[i]), 
			globalAverage_GetValue(t->globalAveMisClassificationCost[i]), t->nSelected[i]);
//		fpOUTFILE << t->strategyId[i] <<", " << globalAverage_GetValue(t->globAveRespTime[i]) << ", "
//		   << expAverage_GetValue(t->expAveRespTime[i]) << ", " << t->nSelected[i] << endl ;

//		outFileDebug(message);
		fprintf(fp, message);
	}

	sprintf(message, "Avg:  %8.4f   %8.4f   %5d   %5d   %5d   %8.4f   %8.4f   %5d\n\n", globalAverage_GetValue(t->overallGlobAveRespTime), 
	   expAverage_GetValue(t->overallExpAveRespTime), t->overallCorrectlyPredicted, t->overallNotPredicted, 
	   t->overallMisPredicted, expAverage_GetValue(t->overallExpAveMisClassCost), 
	   globalAverage_GetValue(t->overallGlobAveMisClassCost), t->overallNSelected);
	//outFileDebug(message);
	fprintf(fp, message);
//	fpOUTFILE << "overall, " << globalAverage_GetValue(t->overallGlobAveRespTime) << ", "
//	   << expAverage_GetValue(t->overallExpAveRespTime) << ", " << t->overallNSelected << endl ;

	fclose(fp);
//	outFileDebug("END strategyTable_Print()\n" );
}


int strategyTable_FindIndex(int strategyId)
 /* Determine the array index corresponding to a given strategyId.
  * This method is internal to this class.
  */
{
	int index = 0 ;  
	int found = 0 ;

	//outFileDebug("BEGIN strategyTable_FindIndex()\n" );

	while (index < t->nStrategies && found == 0)
	{
		if (t->strategyId[index] == strategyId)
			found = 1 ;
		else index++ ;
	}

	//outFileDebug("END strategyTable_FindIndex()\n" );

	if (found == 0)
	{
		ofstream fp(ERRFILE, ios::app);
		fp << "strategyTable_FindIndex() Warning: strategy id (" << strategyId  <<" ) not in table\n" ;
		return -1 ;
	}
	else return index ;
}


void strategyTable_UpdateTable(int strategyId, float respTime) 
 /* Update the statistics for the given strategyId 
  */
{
	int index ;

	//outFileDebug("BEGIN strategyTable_UpdateTable()\n" );

	// determine the array index corresponding to the strategyId 
    index = strategyTable_FindIndex(strategyId) ;  
	if (index == -1) return ;

	// update the global average response time 
	globalAverage_Update(t->globAveRespTime[index], respTime) ;

	// update the exp average response time 
	expAverage_Update(t->expAveRespTime[index], respTime) ;

	// update # times selected 
	t->nSelected[index]++ ;

	// update the overall global average response time 
	globalAverage_Update(t->overallGlobAveRespTime, respTime) ;

	// update the overall exp average response time 
	expAverage_Update(t->overallExpAveRespTime, respTime) ;

	// update the overall global average response time 
	globalAverage_Update(t->overallGlobAveMisClassCost, respTime) ;

	// update the overall exp average response time 
	expAverage_Update(t->overallExpAveMisClassCost, respTime) ;

	// update overall # times selected 
	t->overallNSelected++ ;

	//outFileDebug("END strategyTable_UpdateTable()\n" );
}


void strategyTable_UpdateTable2(float respTime) 
 /* version of strategyTable_UpdateTable() that uses the internally stored strategyId
  */
{
	strategyTable_UpdateTable(t->selectedStrategyId, respTime);
}


void strategyTable_UpdateTable3(float respTime, int correctlyPredicted, int misPredicted, int notPredicted)
{
	int index ;
	float lastMisClassCost ;

	char message[126];

	//outFileDebug("BEGIN strategyTable_UpdateTable3()\n" );

	// determine the array index corresponding to the strategyId 
    index = strategyTable_FindIndex(t->selectedStrategyId) ;  
	if (index == -1) return ;

	// update the global average response time 
	globalAverage_Update(t->globAveRespTime[index], respTime) ;

	// update the exp average response time 
	expAverage_Update(t->expAveRespTime[index], respTime) ;

	// update # times selected 
	t->nSelected[index]++ ;

	// update the overall global average response time 
	globalAverage_Update(t->overallGlobAveRespTime, respTime) ;

	// update the overall exp average response time 
	expAverage_Update(t->overallExpAveRespTime, respTime) ;

	// update overall # times selected 
	t->overallNSelected++ ;

	// P120202 ??? verify Ger
	t->correctlyPredicted[index] += correctlyPredicted ;
	t->notPredicted[index] += notPredicted ;
	t->misPredicted[index] += misPredicted ;

	t->overallCorrectlyPredicted += correctlyPredicted ;
	t->overallNotPredicted += notPredicted ;
	t->overallMisPredicted += misPredicted ;

	lastMisClassCost = strategyTable_calculateMisClassCost(correctlyPredicted, notPredicted, misPredicted) ;
	expAverage_Update(t->expAveMisClassificationCost[index], lastMisClassCost ) ;
	globalAverage_Update(t->globalAveMisClassificationCost[index], lastMisClassCost ) ;

	// update the overall global average response time 
	globalAverage_Update(t->overallGlobAveMisClassCost, lastMisClassCost) ;

	// update the overall exp average response time 
	expAverage_Update(t->overallExpAveMisClassCost, lastMisClassCost) ;

	sprintf(message, "%d\t %d\t %d\t %.6f\t %.6f\t %.6f\n", correctlyPredicted, notPredicted, misPredicted, 
		lastMisClassCost, globalAverage_GetValue(t->globalAveMisClassificationCost[index]), 
		expAverage_GetValue(t->expAveMisClassificationCost[index]) ) ;
/*
	ofstream fpPrintFile;

	switch (t->selectedStrategyId)
	{
		case NO_PREFETCH		: 	fpPrintFile.open("./RunLog/noPref.txt", ios::app);  break ;
		case RANDOM_PREFETCH	: 	fpPrintFile.open("./RunLog/randomPref.txt", ios::app);  break ;
		case DIRECTION_PREFETCH : 	fpPrintFile.open("./RunLog/dirPref.txt", ios::app);  break ;
		case FOCUS_PREFETCH		: 	fpPrintFile.open("./RunLog/focPref.txt", ios::app);  break ;
		default					:	fpPrintFile.open("./RunLog/pref.txt", ios::app);  break ;
	}

	fpPrintFile << message << flush ;
	fpPrintFile.close(); */

	//outFileDebug("END strategyTable_UpdateTable3()\n" );
}

	
int strategyTable_SelectBest(void)
 /* Selects strategy based on SelectionPolicy = Best. 
  * If ties occur, randomly choose among tied strategies. 
  * Output: strategyId of the best strategy.
  * This routine is private to this class.
  */
{
	int idxBest ;
	float bestVal ;
	float tempVal ;
	int i ;
	double random ;
	char message[126] ;

	//outFileDebug("BEGIN strategyTable_SelectBest()\n" );

	if (t->fitnessFunction == STRATEGYTABLE_FITGLOBAVERESPTIME)	
	{
		// if fitness function = global average response time 
		// for response time, the lower the response time, the better 
		idxBest = 0 ;
		bestVal = globalAverage_GetValue(t->globAveRespTime[idxBest]) ;
		for (i = 1; i < t->nStrategies; i++)
		{
			tempVal = globalAverage_GetValue(t->globAveRespTime[i]) ; 
			if (bestVal > tempVal)
			{
				idxBest = i ;
				bestVal = tempVal ;
			}
			else if (bestVal == tempVal)
			{
				// there is a tie. randomly choose between the 2 strategies.
				random = RAND01() ;  // uniform, between [0,1]
				if (random < 0.5)
					idxBest = i ;
			}
		}
	}
	else if (t->fitnessFunction == STRATEGYTABLE_FITEXPAVERESPTIME)
	{
		// if fitness function = exp ave response time.
		// for response time, the lower the response time, the better.
		idxBest = 0 ;
		bestVal = expAverage_GetValue(t->expAveRespTime[idxBest]) ;
		for (i = 1; i < t->nStrategies; i++)
		{
			tempVal = expAverage_GetValue(t->expAveRespTime[i]) ; 
			if (bestVal > tempVal)
			{
				idxBest = i ;
				bestVal = tempVal ;
			}
			else if (bestVal == tempVal)
			{
				// there is a tie. randomly choose between the 2 strategies.
				random = RAND01() ;  // uniform, between [0,1]
				if (random < 0.5)
					idxBest = i ;
			}
		}
	}
	else if (t->fitnessFunction == STRATEGYTABLE_FITEXPAVEMISCLASS_COST)
	{
		// if fitness function = exp ave missclassification cost.
		// for missclassification cost, the lower, the better.
		idxBest = 0 ;
		bestVal = expAverage_GetValue(t->expAveMisClassificationCost[idxBest]) ;
		for (i = 1; i < t->nStrategies; i++)
		{
			tempVal = expAverage_GetValue(t->expAveMisClassificationCost[i]) ; 
			if (bestVal > tempVal)
			{
				idxBest = i ;
				bestVal = tempVal ;
			}
			else if (bestVal == tempVal)
			{
				// there is a tie. randomly choose between the 2 strategies.
				random = RAND01() ;  // uniform, between [0,1]
				if (random < 0.5)
					idxBest = i ;
			}
		}
	}
	else if (t->fitnessFunction == STRATEGYTABLE_FITGLOBALAVEMISCLASS_COST)
	{
		// if fitness function = exp ave missclassification cost.
		// for missclassification cost, the lower, the better.
		idxBest = 0 ;
		bestVal = globalAverage_GetValue(t->globalAveMisClassificationCost[idxBest]) ;
		for (i = 1; i < t->nStrategies; i++)
		{
			tempVal = globalAverage_GetValue(t->globalAveMisClassificationCost[i]) ; 
			if (bestVal > tempVal)
			{
				idxBest = i ;
				bestVal = tempVal ;
			}
			else if (bestVal == tempVal)
			{
				// there is a tie. randomly choose between the 2 strategies.
				random = RAND01() ;  // uniform, between [0,1]
				if (random < 0.5)
					idxBest = i ;
			}
		}
	}
	else
	{
		ofstream fp(ERRFILE, ios::app);
//		FILE *fp = fopen(ERRFILE,"a");
		fp << "strategyTable_SelectBest() Warning: invalid fitness function" << t->fitnessFunction << endl ;
//		fprintf(fp, "strategyTable_SelectBest() Warning: invalid fitness function%d\n", t->fitnessFunction) ;
//		fclose(fp);
		idxBest = 0 ;  // by default, always choose the first strategy
	}

//	strategyTable_Print();

	//outFileDebug("\n===========================================\n");
	sprintf(message, "\n\n@@@@@@@@@@@@@@@@@@@@ Selected Strategy Id: %d\tBest Val=%f\n", idxBest, bestVal );
	//outFileDebug(message) ;
	//outFileDebug("END strategyTable_SelectBest()\n" );

	return t->strategyId[idxBest] ;
}


int strategyTable_SelectProp(void)
 /* Selects strategy based on SelectionPolicy = Proportionate. 
  * This routine is private to this class.
  */
{
	int idxSelect ;
	float sumFit ;
	float tempFit ;
	int i ;
	float random ;
	char message[126];

	//outFileDebug("BEGIN strategyTable_SelectProp()\n" );

	if (t->fitnessFunction == STRATEGYTABLE_FITGLOBAVERESPTIME)	
	{
		// if fitness function = global average response time.
		// for response time, the lower the response time, the better .
		// add 1 to all the values so all values are > 0. otherwise proportionate selection 
		// will ignore the fitness=0 case.

		// add up all values. if low value means better performance, use 1/fitness value
		sumFit = 0.0 ;
		for (i = 0; i < t->nStrategies; i++)
			sumFit = (float) (sumFit + 1.0/(1.0 + globalAverage_GetValue(t->globAveRespTime[i]))) ;
//		printf("sumFit: %f\n", sumFit) ;

		// select strategy proportionately
		random = RAND01() ;
//		printf("random: %f\n", random) ;

		tempFit = 0.0 ;
		for (i = 0; (i < t->nStrategies) && (random > tempFit); i++)
		{
			tempFit = (float) (tempFit + 
				1.0/(sumFit * (1.0 + globalAverage_GetValue(t->globAveRespTime[i])))) ;
//			printf("tempFit: %f\n", tempFit) ;
		}
		idxSelect = i - 1 ;
	}
	else if (t->fitnessFunction == STRATEGYTABLE_FITEXPAVERESPTIME)	
	{
		// if fitness function = exponential average response time.
		// for response time, the lower the response time, the better .
		// add 1 to all the values so all values are > 0. otherwise proportionate selection 
		// will ignore the fitness=0 case.

		// add up all values. if low value means better performance, use 1/fitness value
		sumFit = 0.0 ;
		for (i = 0; i < t->nStrategies; i++)
			sumFit = (float) (sumFit + 1.0/(1.0 + expAverage_GetValue(t->expAveRespTime[i]))) ;
//		printf("sumFit: %f\n", sumFit) ;

		// select strategy proportionately
		random = RAND01() ;
//		printf("random: %f\n", random) ;

		tempFit = 0.0 ;
		for (i = 0; (i < t->nStrategies) && (random > tempFit); i++)
		{
			tempFit = (float) (tempFit + 
				1.0/(sumFit * (1.0 + expAverage_GetValue(t->expAveRespTime[i])))) ;
//			printf("tempFit: %f\n", tempFit) ;
		}
		idxSelect = i - 1 ;
	}
	else if (t->fitnessFunction == STRATEGYTABLE_FITGLOBALAVEMISCLASS_COST)	
	{
		// if fitness function = exponential average misclassification cost.
		// for response time, the lower the response time, the better .
		// add 1 to all the values so all values are > 0. otherwise proportionate selection 
		// will ignore the fitness=0 case.

		// add up all values. if low value means better performance, use 1/fitness value
		sumFit = 0.0 ;
		for (i = 0; i < t->nStrategies; i++)
			sumFit = (float) (sumFit + 1.0/(1.0 + globalAverage_GetValue(t->globalAveMisClassificationCost[i]))) ;
//		printf("sumFit: %f\n", sumFit) ;

		// select strategy proportionately
		random = RAND01() ;
//		printf("random: %f\n", random) ;

		tempFit = 0.0 ;
		for (i = 0; (i < t->nStrategies) && (random > tempFit); i++)
		{
			tempFit = (float) (tempFit + 
				1.0/(sumFit * (1.0 + globalAverage_GetValue(t->globalAveMisClassificationCost[i])))) ;
//			printf("tempFit: %f\n", tempFit) ;
		}
		idxSelect = i - 1 ;
	}
	else if (t->fitnessFunction == STRATEGYTABLE_FITEXPAVEMISCLASS_COST)	
	{
		// if fitness function = exponential average misclassification cost.
		// for misclassification cost, the lower, the better .
		// add 1 to all the values so all values are > 0. otherwise proportionate selection 
		// will ignore the fitness=0 case.

		// add up all values. if low value means better performance, use 1/fitness value
		sumFit = 0.0 ;
		for (i = 0; i < t->nStrategies; i++)
			sumFit = (float) (sumFit + 1.0/(1.0 + expAverage_GetValue(t->expAveMisClassificationCost[i]))) ;
//		printf("sumFit: %f\n", sumFit) ;

		// select strategy proportionately
		random = RAND01() ;
//		printf("random: %f\n", random) ;

		tempFit = 0.0 ;
		for (i = 0; (i < t->nStrategies) && (random > tempFit); i++)
		{
			tempFit = (float) (tempFit + 
				1.0/(sumFit * (1.0 + expAverage_GetValue(t->expAveMisClassificationCost[i])))) ;
//			printf("tempFit: %f\n", tempFit) ;
		}
		idxSelect = i - 1 ;
	}
	else
	{
		ofstream fp(ERRFILE, ios::app);
		fp << "strategyTable_SelectProp() Warning: invalid fitness function" << t->fitnessFunction << endl ;
		idxSelect = 0 ;  // by default, always choose the first strategy
	}

//	strategyTable_Print();

	//outFileDebug("\n===========================================\n");
	sprintf(message, "\n\n@@@@@@@@@@@@@@@@@@@@ Selected Strategy Id: %d\n", idxSelect);
	//outFileDebug(message) ;
	//outFileDebug("END strategyTable_SelectProp()\n" );

	return t->strategyId[idxSelect] ;
}


int strategyTable_SelectStrategy(void)
 /* Selects a strategy based on the Selection Policy and Fitness Function.
  * output: strategyId of selected strategy
  */
{
	int id ;

	if (t->nStrategies == 1)
		id = t->strategyId[0] ;
	else if (t->selectionPolicy == STRATEGYTABLE_POLICYBEST)
		id = strategyTable_SelectBest() ;
	else if (t->selectionPolicy == STRATEGYTABLE_POLICYPROP)
		id = strategyTable_SelectProp() ;
	else		// use default
		id = strategyTable_SelectBest() ;
	
	t->selectedStrategyId = id ;
	return id ;
}


void strategyTable_Test(int testcase)
 /* Tests the entire strategyTable class 
  */
{
/*
	strategyTable t ;
	int selectId ;

	printf("Testing strategyTable class -- begin\n") ;

	if ((t = strategyTable_Create()) == NULL)
		printf("No more memory\n") ;
	else
	{
		//strategyTable_SetFitnessFunction(t, STRATEGYTABLE_FITGLOBAVERESPTIME) ;
		strategyTable_SetFitnessFunction(t, STRATEGYTABLE_FITEXPAVERESPTIME) ;

		//strategyTable_SetSelectionPolicy(t, STRATEGYTABLE_POLICYBEST) ;
		strategyTable_SetSelectionPolicy(t, STRATEGYTABLE_POLICYPROP) ;

		strategyTable_AddStrategy(t, RANDOM_PREFETCH) ;
		strategyTable_AddStrategy(t, DIRECTION_PREFETCH) ;
		strategyTable_AddStrategy(t, FOCUS_PREFETCH) ;
		strategyTable_Print(t) ;

		if (testcase == 0)
		{
			for (int i = 0; i < 10; i++)
			{
				selectId = strategyTable_SelectStrategy(t) ;
				printf("strategy selected: %d\n", selectId) ;
				strategyTable_UpdateTable(t, selectId, (float) 10.0*RAND01()) ;
				strategyTable_Print(t) ;	
			}
		}
		else if (testcase == 1)
		{
			strategyTable_UpdateTable(t, RANDOM_PREFETCH, 1) ;
			strategyTable_Print(t) ;
			printf("strategy selected: %d\n", strategyTable_SelectStrategy(t)) ;

			strategyTable_UpdateTable(t, RANDOM_PREFETCH, 2) ;
			strategyTable_Print(t) ;
			printf("strategy selected: %d\n", strategyTable_SelectStrategy(t)) ;

			strategyTable_UpdateTable(t, RANDOM_PREFETCH, 3) ;
			strategyTable_Print(t) ;
			printf("strategy selected: %d\n", strategyTable_SelectStrategy(t)) ;

			strategyTable_UpdateTable(t, RANDOM_PREFETCH, 4) ;
			strategyTable_Print(t) ;
			printf("strategy selected: %d\n", strategyTable_SelectStrategy(t)) ;

			strategyTable_UpdateTable(t, DIRECTION_PREFETCH, 1) ;
			strategyTable_Print(t) ;
			printf("strategy selected: %d\n", strategyTable_SelectStrategy(t)) ;

			strategyTable_UpdateTable(t, DIRECTION_PREFETCH, 2) ;
			strategyTable_Print(t) ;
			printf("strategy selected: %d\n", strategyTable_SelectStrategy(t)) ;

			strategyTable_UpdateTable(t, DIRECTION_PREFETCH, 3) ;
			strategyTable_Print(t) ;
			printf("strategy selected: %d\n", strategyTable_SelectStrategy(t)) ;

			strategyTable_UpdateTable(t, DIRECTION_PREFETCH, 4) ;
			strategyTable_Print(t) ;
			printf("strategy selected: %d\n", strategyTable_SelectStrategy(t)) ;

			strategyTable_UpdateTable(t, FOCUS_PREFETCH, 1) ;
			strategyTable_Print(t) ;
			printf("strategy selected: %d\n", strategyTable_SelectStrategy(t)) ;

			strategyTable_UpdateTable(t, FOCUS_PREFETCH, 2) ;
			strategyTable_Print(t) ;
			printf("strategy selected: %d\n", strategyTable_SelectStrategy(t)) ;

			strategyTable_UpdateTable(t, FOCUS_PREFETCH, 3) ;
			strategyTable_Print(t) ;
			printf("strategy selected: %d\n", strategyTable_SelectStrategy(t)) ;

			strategyTable_UpdateTable(t, FOCUS_PREFETCH, 4) ;
			strategyTable_Print(t) ;
			printf("strategy selected: %d\n", strategyTable_SelectStrategy(t)) ;
		}
	}

	strategyTable_Destroy(t) ;

	printf("Testing strategyTable class -- done\n") ;
	return ;
*/
}

void strategyTable_SelectPrefetchers(void)
/* Hard code the table with the existing prefetchers
 */
{
	//outFileDebug("BEGIN strategyTable_SelectPrefetchers()\n" );

	strategyTable_Create();

	//strategyTable_SetFitnessFunction(STRATEGYTABLE_FITGLOBAVERESPTIME) ;
	//strategyTable_SetFitnessFunction(STRATEGYTABLE_FITEXPAVERESPTIME) ;
	strategyTable_SetFitnessFunction(STRATEGYTABLE_FITEXPAVEMISCLASS_COST) ;
//	strategyTable_SetFitnessFunction(STRATEGYTABLE_FITGLOBALAVEMISCLASS_COST) ;

	strategyTable_SetSelectionPolicy(STRATEGYTABLE_POLICYBEST) ;
	//strategyTable_SetSelectionPolicy(STRATEGYTABLE_POLICYPROP) ;

	strategyTable_AddStrategy(NO_PREFETCH) ;
	strategyTable_AddStrategy(RANDOM_PREFETCH) ;
	strategyTable_AddStrategy(DIRECTION_PREFETCH) ;
//	strategyTable_AddStrategy(FOCUS_PREFETCH) ;
//	strategyTable_Print() ;

	//outFileDebug("END strategyTable_SelectPrefetchers()\n" );

}

int strategyTable_GetSelectedStrategy(void)
 /* Return the internally stored selected strategy id
  */
{
	return t->selectedStrategyId ;
}
 
