// MetaData.cpp: implementation of the MetaData class.
//
//////////////////////////////////////////////////////////////////////

#pragma warning(disable:4786)

#include <iostream>
#include <fstream>
#include <assert.h>

#include <SAX/InputSource.h>
#include <SAX/helpers/DefaultHandler.h>

#include <meta/Meta.h>
#include <meta/MetaData.h>
#include <meta/Dimension.h>

#include <meta/MetaHelper.h>

//////////////////////////////////////////////////////////////////////

class MetaData::MetaDataHandler : public SAX::DefaultHandler
{
public:
	MetaDataHandler(MetaData *meta);
	~MetaDataHandler();
	
	SAX::InputSource input;
	
	////////////////////////////////////////////////////
	// DefaultHandler
	virtual void startElement(const std::string& namespaceURI, 
		const std::string& localName,
		const std::string& qName, 
		const AttributesT& atts);
	virtual void endElement(const std::string& namespaceURI, 
		const std::string& localName,
		const std::string& qName);

private:
	MetaData *mMeta;
};

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

MetaData::MetaData()
{
	mHandler = new MetaDataHandler(this);

	// register the types
	Integer::getTypeId();
	Real::getTypeId();
	Word::getTypeId();
	String::getTypeId();
	URI::getTypeId();
}

MetaData::~MetaData()
{
	DimVecIter i = mDimensions.begin();
	for ( ; i != mDimensions.end(); ++i) {
		delete (Dimension*) (*i);
	}
	mDimensions.clear();

	delete mHandler;
}

bool MetaData::unmarshal(std::istream &is, ISerializable *parent)
{
	METAHELPER->pushHandler(mHandler);

	if (parent == 0) {
		mHandler->input.setByteStream(is);
		
		if(mHandler->input.getByteStream() == 0)
			return false;
		METAHELPER->getParser().parse(mHandler->input);
	}
	
	return true;
}

bool MetaData::marshal(std::ostream& os, ISerializable *parent)
{
	os << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << std::endl;
	os << "<okcmeta>" << std::endl;

	os << "<dimensions value=\"" << mDimensionsCount << "\" />" << std::endl;
	os << "<datapoints value=\"" << mDataPointsCount << "\" />" << std::endl;

	DimVecCIter i = mDimensions.begin();
	for ( ; i != mDimensions.end(); ++i) {
		IDimension *dim = *i;
		dim->marshal(os, this);
	}

	os << "</okcmeta>" << std::endl;
	
	return true;
}

int MetaData::getDimensionCount() const
{
	return mDimensionsCount;
}

void MetaData::setDimensionCount(int dimensions)
{
	mDimensionsCount = dimensions;
}

int MetaData::getDataPointCount() const
{
	return mDataPointsCount;
}

void MetaData::setDataPointCount(int datapoints)
{
	mDataPointsCount = datapoints;
}

IDimension *MetaData::getDimension(int dim) const
{
	return mDimensions.at(dim);
}

IDimension *MetaData::getDimension(const char* dimName) const
{
	IDimension *dimension = NULL;
	std::vector<IDimension*>::const_iterator i = mDimensions.begin();
	for ( ; i != mDimensions.end(); ++i) {
		dimension = *i;
		if (!strcmp(dimension->getName(), dimName)) {
			break;
		}
	}
	return dimension;
}

IDimension* MetaData::createDimension(const char* dimName,Type dimType) const
{
	Dimension *dimension = new Dimension();
	dimension->setName(dimName);
	dimension->setType(dimType);
	return dimension;
}

void MetaData::addDimension(IDimension *dimension)
{
	mDimensions.push_back(dimension);
}

//////////////////////////////////////////////////////////////////////////
// MetaDataHandler

MetaData::MetaDataHandler::MetaDataHandler(MetaData *meta)
{
	mMeta = meta;
//	SAX::FeatureNames<std::string> fNames;
//	try {
//		METAHELPER->getParser().setFeature(fNames.namespaces, true);
//		METAHELPER->getParser().setFeature(fNames.namespace_prefixes, true);
//	} catch (...) {
//		// Empty		
//	}
}

MetaData::MetaDataHandler::~MetaDataHandler()
{
}

/* virtual */
void MetaData::MetaDataHandler::startElement(const std::string& namespaceURI, 
											 const std::string& localName,
											 const std::string& qName, 
											 const SAX::Attributes& atts)
{
	if(localName.length()) {
//		std::cerr << "Start Elem: " << localName << std::endl;
		if (localName == "dimensions") {

			mMeta->mDimensionsCount = atoi(atts.getValue("value").c_str());

		} else if (localName == "datapoints") {

			mMeta->mDataPointsCount = atoi(atts.getValue("value").c_str());

		} else if (localName == "simpleType") {

			std::string strName = atts.getValue("name");
			std::string strBaseType = atts.getValue("base");

			Type type = Type::fromName(strBaseType.c_str());
			if (type.isDerivedFrom(BasicTypeValue::getTypeId())) {
				SimpleValue *value = new SimpleValue(strName, type);
				SimpleValue::getTypeId(value);	// register this new type
				value->unmarshal(*input.getByteStream(), mMeta);
			}


		} else if (localName == "dimension") {

			std::string strName = atts.getValue("name");
			std::string strType = atts.getValue("type");

			Dimension *dimension = new Dimension();
			mMeta->mDimensions.push_back(dimension);
			dimension->unmarshal(*input.getByteStream(), mMeta);
			dimension->setName(strName.c_str());
			dimension->setType(Type::fromName(strType.c_str()));
			
		}
	}
}

/* virtual */
void MetaData::MetaDataHandler::endElement(const std::string& namespaceURI, 
										   const std::string& localName,
										   const std::string& qName)
{
	if (localName.length()) {
//		std::cerr << "End Elem: " << localName << std::endl;
		if (localName == "okcmeta") {
			assert (this == METAHELPER->popHandler());
		}
	}
}
